from StringIO import StringIO
import datetime
import dircache
import unittest
import tempfile
import shutil
import new
import sys
import os

from MoinMoin.config.multiconfig import DefaultConfig, _config_cache
from MoinMoin.formatter.text_plain import Formatter as TextFormatter
from MoinMoin.request import request_cli
from MoinMoin.user import User
from MoinMoin.PageEditor import PageEditor
from MoinMoin.Page import Page


Calc_path = os.path.join(os.path.dirname(__file__), "Calc.py")
if not os.path.isfile(Calc_path):
    sys.exit("error: %s not found" % Calc_path)


class TestConfig(DefaultConfig):

    data_dir = None # Set by setUp().
    data_underlay_dir = None
    show_hosts = 1
    sitename = logo_string = "Calc Test Wiki"


class CalcTest(unittest.TestCase):

    def join(self, *args):
        return os.path.join(self.wiki_path, *args)

    def makedirs(self, *args):
        path = self.join(*args)
        if not os.path.isdir(path):
            return os.makedirs(path)

    def touch(self, *args):
        path = self.join(*args)
        parent_path = os.path.dirname(path)
        if not os.path.isdir(parent_path):
            os.makedirs(parent_path)
        open(path, "w").close()

    def write(self, content, *args):
        path = self.join(*args)
        os.makedirs(os.path.dirname(path))
        file = open(path, "w")
        file.write(content)
        file.close()

    def setUp(self):
        self.wiki_path = tempfile.mkdtemp()

        TestConfig.data_dir = self.join("data")

        self.touch("wikiconfig.py")
        self.touch("data/__init__.py")
        self.touch("data/pages/__init__.py")
        self.touch("data/plugin/__init__.py")
        self.write("modules = ['Calc']",
                   "data/plugin/macro/__init__.py")

        os.link(Calc_path, self.join("data/plugin/macro/Calc.py"))

        self.wikiconfig = sys.modules["wikiconfig"] = new.module("wikiconfig")
        self.wikiconfig.__file__ = self.join("wikiconfig.py")
        self.wikiconfig.Config = TestConfig

        self.request = request_cli.Request()
        self.request.form = self.request.args = self.request.setup_args()
        self.request.user = User(self.request)

    def tearDown(self):
        shutil.rmtree(self.wiki_path)
        del sys.modules["wikiconfig"]
        _config_cache.clear()

    def render(self, page_name, content):
        content = unicode(content)
        formatter = TextFormatter(self.request, is_included=True)
        page = Page(self.request, page_name, formatter=formatter)
        page.set_raw_body(content)
        sio = StringIO()
        self.request.redirect(sio)
        try:
            page.send_page(content_only=1)
            return sio.getvalue().strip()
        finally:
            self.request.redirect()

    def save(self, page_name, content):
        page = PageEditor(self.request, page_name, do_editor_backup=0)
        page.saveText(content, 0)

        # Moin uses dircache, and it was returning obsolete data to the
        # getPageList() call.
        dircache.reset()

    def check(self, content, rendered_body):
        self.assertEquals(self.render("TestPage", content), rendered_body)

    def test_one(self):
        self.check("<<Calc(1)>>", "1")

    def test_neg(self):
        self.check("<<Calc(-1)>>", "-1")

    def test_add(self):
        self.check("<<Calc(1+2)>>", "3")

    def test_sub(self):
        self.check("<<Calc(1-2)>>", "-1")

    def test_mul(self):
        self.check("<<Calc(2*3)>>", "6")

    def test_div(self):
        self.check("<<Calc(3/2)>>", "1.5")

    def test_pow(self):
        self.check("<<Calc(2**3**4)>>", "2417851639229258349412352")

    def test_negative_pow(self):
        self.check("<<Calc(2**-1)>>", "0.5")

    def test_pow_large(self):
        self.check("<<Calc(10**10000)>>",
                   "1.000000000000000000000000000E+10000")

    def test_pow_badly_large(self):
        self.check("<<Calc(10**100000)>>",
                   "*** <<Calc["
                   "Magnitude error in expression: 10**100000"
                   "]>> ***")

        self.check("<<Calc(10**1000**1000**1000)>>",
                   "*** <<Calc["
                   "Magnitude error in expression: 10**1000**1000**1000"
                   "]>> ***")

    def test_pow_not_a_number(self):
        self.check("<<Calc(extract_list=\d+, define=l)>>1<<Calc(10**l)>>",
                   "1\n\n*** <<Calc["
                   "Exponent isn't a number in expression: 10**l"
                   "]>> ***")

    def test_precedence(self):
        self.check("<<Calc(1+2*3+4**5*6**7*8+9)>>", "2293235728")

    def test_parenthesis(self):
        self.check("<<Calc((1+2)*(3+4))>>", "21")

    def test_bad_expression(self):
        self.check("<<Calc(&*())>>",
                   "*** <<Calc[Parser error in expression: &*()]>> ***")

    def test_keyword_expression(self):
        self.check("<<Calc(expression=1+2)>>", "3")

    def test_keyword_and_positional_expression(self):
        self.check("<<Calc(1, expression=2)>>",
                   "*** <<Calc["
                   "Multiple expression arguments (positional and keyword)"
                   "]>> ***")

    def test_define(self):
        self.check("<<Calc(1+2, define=name1)>>|<<Calc(name1+1)>>", "|4")

    def test_define_function(self):
        self.check("<<Calc(1+name1, define=name2())>>|"
                   "<<Calc(2, define=name1)>>|"
                   "<<Calc(name2()+1)>>", "||4")

    def test_define_function_with_vars(self):
        self.check("<<Calc((1+a)*b, define=name(a;b))>>|<<Calc(name(2;3))>>",
                   "|9")

    def test_define_function_infinite_recursion(self):
        self.check("<<Calc(func(), define=func())>>|<<Calc(func())>>",
                "|\n\n*** <<Calc["
                "Maximum function depth reached in expression: func()"
                "]>> ***")

    def test_define_function_wrong_number_of_arguments(self):
        self.check("<<Calc(1, define=func(a;b;c))>>|<<Calc(func(1;2))>>",
                "|\n\n*** <<Calc["
                "func(a;b;c) got 2 arguments in expression: func(1;2)"
                "]>> ***")

    def test_show(self):
        self.check("<<Calc(1+2, define=name1, show=1)>>|<<Calc(name1+1)>>",
                   "3|4")

    def test_undefined_variable(self):
        self.check("<<Calc(name1+name2)>>",
                   "*** <<Calc["
                   "'name1' variable not defined in expression: name1+name2"
                   "]>> ***")

    def test_import_from(self):
        self.save("ImportedPage", "<<Calc(1, define=name)>>")
        self.check("<<Calc(name, import_from=ImportedPage)>>", "1")

    def test_import_from_pattern(self):
        self.save("ImportedPage", "<<Calc(1, define=name)>>")
        self.check("<<Calc(name, import_from=^[I]mportedPage)>>", "1")

    def test_import_from_bad_pattern(self):
        self.save("ImportedPage", "<<Calc(1, define=name)>>")
        self.check("<<Calc(name, import_from=[I]mportedPage)>>",
                   "*** <<Calc[Page '[I]mportedPage' doesn't exist]>> ***")

    def test_import_list_from(self):
        self.save("ImportedPage1", "<<Calc(1, define=name)>>")
        self.save("ImportedPage2", "<<Calc(2, define=name)>>")
        self.save("ImportedPage3", "<<Calc(3, define=name)>>")
        self.check("<<Calc(name, import_list_from=^ImportedPage[13])>>",
                   "[1, 3]")

    def test_import_function(self):
        self.save("ImportedPage", "<<Calc(1, define=func())>>")
        self.check("<<Calc(func(), import_from=ImportedPage)>>", "1")

    def test_extract_found_nothing(self):
        self.check("<<Calc(extract=\d)>>",
                   "*** <<Calc["
                   "Extract pattern found nothing: \d"
                   "]>> ***")

    def test_extract_prevents_self_extraction(self):
        self.check("<<Calc(extract=[1-3])>>|2", "2|2")

    def test_extract_value(self):
        self.check("Hello: 123 <<Calc(extract=\d+)>>",
                   "Hello: 123 123")

    def test_extract_value_before(self):
        self.check("<<Calc(extract=\d+)>> Hello: 123",
                   "123 Hello: 123")

    def test_extract_value_with_group(self):
        self.check("Hello: 123 <<Calc(extract=Hello: (\d+))>>",
                   "Hello: 123 123")

    def test_extract_value_multiline(self):
        self.check("1\n2\n3|<<Calc(extract=1.*(\d).*3)>>", "1 2 3|2")

    def test_extract_value_with_too_many_groups(self):
        self.check("12|<<Calc(extract=(1)(2))>>",
                   "12|\n\n*** <<Calc["
                   "More than one group in extract pattern: (1)(2)"
                   "]>> ***")

    def test_extract_expression(self):
        self.check("Hello: 123+1=<<Calc(extract=[+\d]+)>>",
                   "Hello: 123+1=124")

    def test_extract_expression_with_variable(self):
        self.check("<<Calc(1,define=one)>>Hello: 123+one="
                   "<<Calc(extract=( [+\w]+))>>",
                   "Hello: 123+one=124")

    def test_extract_from(self):
        self.save("ExtractPage", ">>123+4<<")
        self.check("<<Calc(extract=>>([+\d]+)<<, extract_from=ExtractPage)>>",
                   "127")

    def test_extract_from_pattern(self):
        self.save("ExtractPage", ">>123+4<<")
        self.check("<<Calc(extract=>>([+\d]+)<<, extract_from=^[E]xtractPage)>>",
                   "127")

    def test_extract_from_unexistent_page(self):
        self.save("ExtractPage", "1")
        self.check("<<Calc(extract=1, extract_from=[E]xtractPage)>>",
                   "*** <<Calc[Page '[E]xtractPage' doesn't exist]>> ***")

    def test_extract_list(self):
        self.check("1 2 3=<<Calc(extract_list=\d+)>>",
                   "1 2 3=[1, 2, 3]")

    def test_extract_list_from(self):
        self.save("ExtractPage", "1 2 3")
        self.check("<<Calc(extract_list=\d, extract_from=ExtractPage)>>",
                   "[1, 2, 3]")

    def test_extract_list_from_pattern(self):
        self.save("ExtractPage1", "1 2 3")
        self.save("ExtractPage2", "4 5 6")
        self.save("ExtractPage3", "7 8 9")
        self.check("<<Calc(extract_list=\d, extract_from=^[E]xtractPage[13])>>",
                   "[1, 2, 3, 7, 8, 9]")

    def test_extract_list_from_unexistent_page(self):
        self.save("ExtractPage", "1 2 3")
        self.check("<<Calc(extract_list=\d, extract_from=[E]xtractPage)>>",
                   "*** <<Calc[Page '[E]xtractPage' doesn't exist]>> ***")

    def test_extract_mimetype(self):
        self.check(r"'''123+4'''|"
                   r"<<Calc(extract=\*([+\d]+)\*, "
                   r"extract_mimetype=text/plain)>>",
                   r"*123+4*|127")

    def test_extract_mimetype_unexistent(self):
        self.check("<<Calc(extract=foo, extract_mimetype=unexistent)>>",
                   "*** <<Calc[Mimetype 'unexistent' not supported]>> ***")

    def test_extract_without_mimetype_wont_render_body(self):
        self.check(r"'''123+4'''|<<Calc(extract='''([+\d]+)''')>>",
                   r"*123+4*|127")

    def test_extract_within(self):
        self.check("123456|<<Calc(extract=\d+, extract_within=2.*5)>>",
                   "123456|2345")

    def test_extract_within_with_group(self):
        self.check("123456|<<Calc(extract=\d+, extract_within=2(.*)5)>>",
                   "123456|34")

    def test_extract_within_matches_nothing(self):
        self.check("<<Calc(extract=\d+, extract_within=123)>>",
                   "*** <<Calc[extract_within found nothing: 123]>> ***")

    def test_extract_within_multiline(self):
        self.check("1\n2\n3|<<Calc(extract=\d+, extract_within=1(.*)3)>>",
                   "1 2 3|2")

    def test_extract_within_with_bad_expression(self):
        self.check("<<Calc(extract=\d+, extract_within=???)>>",
                   "*** <<Calc[Error in extract_within pattern: ???]>> ***")

    def test_extract_within_with_too_many_groups(self):
        self.check("12|<<Calc(extract=\d+, extract_within=(1)(2))>>",
                   "12|\n\n*** <<Calc["
                   "More than one group in extract_within pattern: (1)(2)"
                   "]>> ***")

    def test_string(self):
        self.check("<<Calc('Hello')>>", "Hello")

    def test_string_if(self):
        self.check("<<Calc(IF(1;'True';'False'))>>", "True")
        self.check("<<Calc(IF(0;'True';'False'))>>", "False")

    def test_string_concatenation(self):
        self.check("<<Calc('Hello'+' there!')>>", "Hello there!")

    def test_if_true(self):
        self.check("<<Calc(if(1;2;3/0))>>", "2")

    def test_if_false(self):
        self.check("<<Calc(if(0;1/0;3))>>", "3")

    def test_if_true_with_bad_syntax_in_skip_branch(self):
        self.check("<<Calc(if(1;2;/))>>",
                   "*** <<Calc[Parser error in expression: if(1;2;/)]>> ***")

    def test_if_false_with_bad_syntax_in_skip_branch(self):
        self.check("<<Calc(if(0;/;1))>>",
                   "*** <<Calc[Parser error in expression: if(0;/;1)]>> ***")

    def test_if_without_then_true(self):
        self.check("<<Calc(if(4;;5))>>", "")

    def test_if_without_then_false(self):
        self.check("<<Calc(if(0;;6))>>", "6")

    def test_if_without_else_true(self):
        self.check("<<Calc(if(4;5;))>>", "5")

    def test_if_without_else_false(self):
        self.check("<<Calc(if(0;6;))>>", "")

    def test_if_without_either_branch(self):
        self.check("<<Calc(if(1;;))>>", "")

    def test_if_uppercased(self):
        self.check("<<Calc(IF(1;2;3))>>", "2")

    def test_bad_func(self):
        self.check("<<Calc(badfunc(1))>>", 
                   "*** <<Calc["
                   "'badfunc' is not a function in expression: badfunc(1)"
                   "]>> ***")

    def test_func_uppercased(self):
        self.check("<<Calc(extract_list=\d+, define=l)>>1 2 3="
                   "<<Calc(SUM(l))>>", "1 2 3=6")

    def test_func_sum_with_list(self):
        self.check("<<Calc(extract_list=\d+, define=l)>>1 2 3="
                   "<<Calc(sum(l))>>", "1 2 3=6")

    def test_func_sum_with_values(self):
        self.check("<<Calc(sum(1;2;3))>>", "6")

    def test_func_sum_with_bad_args(self):
        self.check("<<Calc(sum())>>",
                   "*** <<Calc[sum() needs 1 or more arguments and got 0 in "
                   "expression: sum()]>> ***")

    def test_func_min(self):
        self.check("<<Calc(extract_list=\d+, define=l)>>1 2 3="
                   "<<Calc(min(l))>>", "1 2 3=1")

    def test_func_max(self):
        self.check("<<Calc(extract_list=\d+, define=l)>>1 2 3="
                   "<<Calc(max(l))>>", "1 2 3=3")

    def test_func_average(self):
        self.check("<<Calc(extract_list=\d+, define=l)>>1 2 3="
                   "<<Calc(average(l))>>", "1 2 3=2")

    def test_func_len(self):
        self.check("<<Calc(extract_list=\d+, define=l)>>1 2 1="
                   "<<Calc(len(l))>>", "1 2 1=3")

    #def test_func_sum_min_max_average_len_with_non_list_arg(self):
    #    for func in "sum min max average len".split():
    #        self.check("<<Calc(%s(1))>>" % func,
    #                   "*** <<Calc["
    #                   "%s() argument is not a list in expression: %s(1)"
    #                   "]>> ***" % (func, func))

    def test_date(self):
        self.check("<<Calc(date(1977;5;4))>>", "1977-05-04")

    def test_date_bad_args(self):
        self.check("<<Calc(date(1977;4))>>",
                   "*** <<Calc[date() needs 3 arguments and got 2 "
                   "in expression: date(1977;4)]>> ***")

    def test_day(self):
        self.check("<<Calc(day(date(1977;5;4)))>>", "4")

    def test_days(self):
        self.check("<<Calc(days(date(1977;5;4);date(1977;5;7)))>>", "-3")

    def test_today(self):
        self.check("<<Calc(today())>>", str(datetime.date.today()))

    def test_weekday(self):
        self.check("<<Calc(weekday(date(2008;3;22)))>>", "7")
        self.check("<<Calc(weekday(date(2008;3;23)))>>", "1")
        self.check("<<Calc(weekday(date(2008;3;24)))>>", "2")

    def test_weekday_1(self):
        self.check("<<Calc(weekday(date(2008;3;22); 1))>>", "7")
        self.check("<<Calc(weekday(date(2008;3;23); 1))>>", "1")
        self.check("<<Calc(weekday(date(2008;3;24); 1))>>", "2")

    def test_weekday_2(self):
        self.check("<<Calc(weekday(date(2008;3;22); 2))>>", "6")
        self.check("<<Calc(weekday(date(2008;3;23); 2))>>", "7")
        self.check("<<Calc(weekday(date(2008;3;24); 2))>>", "1")

    def test_weekday_3(self):
        self.check("<<Calc(weekday(date(2008;3;22); 3))>>", "5")
        self.check("<<Calc(weekday(date(2008;3;23); 3))>>", "6")
        self.check("<<Calc(weekday(date(2008;3;24); 3))>>", "0")

    def test_choose(self):
        self.check("<<Calc(choose(1;'A';'B';'C'))>>", "A")
        self.check("<<Calc(choose(2;'A';'B';'C'))>>", "B")
        self.check("<<Calc(choose(3;'A';'B';'C'))>>", "C")

    def test_choose_evaluates_lazily(self):
        self.check("<<Calc(choose(1;'A';1/0))>>", "A")

    def test_choose_empty(self):
        self.check("<<Calc(choose(2;;;;))>>", "")

    def test_choose_bad_args(self):
        self.check("<<Calc(choose(0))>>",
                   "*** <<Calc["
                   "choose() needs 2 or more arguments and got 1 in expression:"
                   " choose(0)]>> ***")

    def test_choose_bad_index(self):
        self.check("<<Calc(choose(2;'A'))>>",
                   "*** <<Calc["
                   "No element 2 for choose() in expression: choose(2;'A')"
                   "]>> ***")
        self.check("<<Calc(choose(0;'A'))>>",
                   "*** <<Calc["
                   "No element 0 for choose() in expression: choose(0;'A')"
                   "]>> ***")


if __name__ == "__main__":
    if not os.path.isfile(Calc_path):
        sys.exit("error: Calc.py not found")
    unittest.main()
